/* PIKA - Photo and Image Kooker Application
 * a rebranding of The GNU Image Manipulation Program (created with heckimp)
 * A derived work which may be trivial. However, any changes may be (C)2023 by Aldercone Studio
 *
 * Original copyright, applying to most contents (license remains unchanged): 
 * Copyright (C) 1995 Spencer Kimball and Peter Mattis
 *
 * pikapickablepopup.c
 * Copyright (C) 2014 Michael Natterer <mitch@gimp.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "config.h"

#include <gegl.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>

#include "libpikabase/pikabase.h"
#include "libpikawidgets/pikawidgets.h"

#include "widgets-types.h"

#include "core/pika.h"
#include "core/pikacontext.h"
#include "core/pikaimage.h"
#include "core/pikapickable.h"
#include "core/pikaviewable.h"

#include "pikacontainertreeview.h"
#include "pikacontainerview.h"
#include "pikapickablechooser.h"
#include "pikapickablepopup.h"
#include "pikaviewrenderer.h"

#include "pika-intl.h"


enum
{
  PROP_0,
  PROP_CONTEXT,
  PROP_PICKABLE,
  PROP_VIEW_SIZE,
  PROP_VIEW_BORDER_WIDTH
};

struct _PikaPickablePopupPrivate
{
  PikaContext  *context;

  gint          view_size;
  gint          view_border_width;

  GtkWidget    *chooser;

  GtkWidget    *image_view;
  GtkWidget    *layer_view;
  GtkWidget    *channel_view;
  GtkWidget    *layer_label;
};


static void   pika_pickable_popup_constructed     (GObject             *object);
static void   pika_pickable_popup_finalize        (GObject             *object);
static void   pika_pickable_popup_set_property    (GObject             *object,
                                                   guint                property_id,
                                                   const GValue        *value,
                                                   GParamSpec          *pspec);
static void   pika_pickable_popup_get_property    (GObject             *object,
                                                   guint                property_id,
                                                   GValue              *value,
                                                   GParamSpec          *pspec);

static void   pika_pickable_popup_activate        (PikaPickablePopup   *popup);
static void   pika_pickable_popup_notify_pickable (PikaPickablePopup   *popup);


G_DEFINE_TYPE_WITH_PRIVATE (PikaPickablePopup, pika_pickable_popup,
                            PIKA_TYPE_POPUP)

#define parent_class pika_pickable_popup_parent_class


static void
pika_pickable_popup_class_init (PikaPickablePopupClass *klass)
{
  GObjectClass *object_class = G_OBJECT_CLASS (klass);

  object_class->constructed  = pika_pickable_popup_constructed;
  object_class->finalize     = pika_pickable_popup_finalize;
  object_class->get_property = pika_pickable_popup_get_property;
  object_class->set_property = pika_pickable_popup_set_property;

  g_object_class_install_property (object_class, PROP_CONTEXT,
                                   g_param_spec_object ("context",
                                                        NULL, NULL,
                                                        PIKA_TYPE_CONTEXT,
                                                        PIKA_PARAM_READWRITE |
                                                        G_PARAM_CONSTRUCT_ONLY));

  g_object_class_install_property (object_class, PROP_PICKABLE,
                                   g_param_spec_object ("pickable",
                                                        NULL, NULL,
                                                        PIKA_TYPE_PICKABLE,
                                                        PIKA_PARAM_READWRITE));

  g_object_class_install_property (object_class, PROP_VIEW_SIZE,
                                   g_param_spec_int ("view-size",
                                                     NULL, NULL,
                                                     1, PIKA_VIEWABLE_MAX_PREVIEW_SIZE,
                                                     PIKA_VIEW_SIZE_MEDIUM,
                                                     PIKA_PARAM_READWRITE |
                                                     G_PARAM_CONSTRUCT));

  g_object_class_install_property (object_class, PROP_VIEW_BORDER_WIDTH,
                                   g_param_spec_int ("view-border-width",
                                                     NULL, NULL,
                                                     0,
                                                     PIKA_VIEW_MAX_BORDER_WIDTH,
                                                     1,
                                                     PIKA_PARAM_READWRITE |
                                                     G_PARAM_CONSTRUCT));
}

static void
pika_pickable_popup_init (PikaPickablePopup *popup)
{
  popup->priv = pika_pickable_popup_get_instance_private (popup);

  popup->priv->view_size         = PIKA_VIEW_SIZE_SMALL;
  popup->priv->view_border_width = 1;

  gtk_window_set_resizable (GTK_WINDOW (popup), FALSE);
}

static void
pika_pickable_popup_constructed (GObject *object)
{
  PikaPickablePopup *popup = PIKA_PICKABLE_POPUP (object);

  G_OBJECT_CLASS (parent_class)->constructed (object);

  pika_assert (PIKA_IS_CONTEXT (popup->priv->context));

  popup->priv->chooser = pika_pickable_chooser_new (popup->priv->context, PIKA_TYPE_PICKABLE,
                                                    popup->priv->view_size,
                                                    popup->priv->view_border_width);
  gtk_container_add (GTK_CONTAINER (popup), popup->priv->chooser);
  g_signal_connect_swapped (popup->priv->chooser, "notify::pickable",
                            G_CALLBACK (pika_pickable_popup_notify_pickable),
                            popup);
  g_signal_connect_swapped (popup->priv->chooser, "activate",
                            G_CALLBACK (pika_pickable_popup_activate),
                            popup);
  gtk_widget_show (popup->priv->chooser);
}

static void
pika_pickable_popup_finalize (GObject *object)
{
  PikaPickablePopup *popup = PIKA_PICKABLE_POPUP (object);

  g_clear_object (&popup->priv->context);

  G_OBJECT_CLASS (parent_class)->finalize (object);
}

static void
pika_pickable_popup_set_property (GObject      *object,
                                  guint         property_id,
                                  const GValue *value,
                                  GParamSpec   *pspec)
{
  PikaPickablePopup *popup = PIKA_PICKABLE_POPUP (object);

  switch (property_id)
    {
    case PROP_CONTEXT:
      if (popup->priv->context)
        g_object_unref (popup->priv->context);
      popup->priv->context = g_value_dup_object (value);
      break;

    case PROP_PICKABLE:
      pika_pickable_chooser_set_pickable (PIKA_PICKABLE_CHOOSER (popup->priv->chooser),
                                          g_value_get_object (value));
      break;

    case PROP_VIEW_SIZE:
      popup->priv->view_size = g_value_get_int (value);
      break;

    case PROP_VIEW_BORDER_WIDTH:
      popup->priv->view_border_width = g_value_get_int (value);
      break;

    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

static void
pika_pickable_popup_get_property (GObject    *object,
                                  guint       property_id,
                                  GValue     *value,
                                  GParamSpec *pspec)
{
  PikaPickablePopup *popup = PIKA_PICKABLE_POPUP (object);

  switch (property_id)
    {
    case PROP_CONTEXT:
      g_value_set_object (value, popup->priv->context);
      break;

    case PROP_PICKABLE:
      g_value_set_object (value, pika_pickable_popup_get_pickable (popup));
      break;

    case PROP_VIEW_SIZE:
      g_value_set_int (value, popup->priv->view_size);
      break;

    case PROP_VIEW_BORDER_WIDTH:
      g_value_set_int (value, popup->priv->view_border_width);
      break;

   default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}


/* Public functions */

GtkWidget *
pika_pickable_popup_new (PikaContext *context,
                         gint         view_size,
                         gint         view_border_width)
{
  g_return_val_if_fail (PIKA_IS_CONTEXT (context), NULL);
  g_return_val_if_fail (view_size >  0 &&
                        view_size <= PIKA_VIEWABLE_MAX_POPUP_SIZE, NULL);
  g_return_val_if_fail (view_border_width >= 0 &&
                        view_border_width <= PIKA_VIEW_MAX_BORDER_WIDTH,
                        NULL);

  return g_object_new (PIKA_TYPE_PICKABLE_POPUP,
                       "type",              GTK_WINDOW_POPUP,
                       "context",           context,
                       "view-size",         view_size,
                       "view-border-width", view_border_width,
                       NULL);
}

PikaPickable *
pika_pickable_popup_get_pickable (PikaPickablePopup *popup)
{
  g_return_val_if_fail (PIKA_IS_PICKABLE_POPUP (popup), NULL);

  return pika_pickable_chooser_get_pickable (PIKA_PICKABLE_CHOOSER (popup->priv->chooser));
}


/* Private functions */

static void
pika_pickable_popup_activate (PikaPickablePopup *popup)
{
  g_signal_emit_by_name (popup, "confirm");
}

static void
pika_pickable_popup_notify_pickable (PikaPickablePopup *popup)
{
  g_object_notify (G_OBJECT (popup), "pickable");
}
